from datetime import datetime
from pydantic import BaseModel
from .Extras import Post, Giver

# https://osu.ppy.sh/docs/index.html#kudosuhistory
class KudosuHistory(BaseModel):
    id: int
    action: str
    account: int|None
    model: str
    created_at: datetime
    giver: Giver | None
    post: Post

    class Config:
        arbitrary_types_allowed = True
        json_encoders = {
            datetime: str,
		}



from datetime import datetime
from typing import Any
from pydantic import BaseModel, Field
from .Extras import Availability, Nominations, Hype
from .User import UserCompact
from .Covers import Covers

# https://osu.ppy.sh/docs/index.html#beatmapsetcompact
class BeatmapsetCompact(BaseModel):
    artist: str
    artist_unicode: str
    covers: Covers
    creator: str
    favorite_count: str = Field(alias="favourite_count")
    id: int
    nsfw: bool
    play_count: int
    preview_url: str
    source: str
    status: str
    title: str
    title_unicode: str
    user_id: int
    video: bool
    
    # optionals
    beatmaps: list[Any] | None
    converts: Any | None
    current_user_attributes: Any | None
    description: Any | None
    discussions: Any | None
    events: Any | None
    genre: Any | None
    has_favourited: bool | None
    language: Any | None
    nominations: Any | None
    ratings: list[int] | None
    recent_favourites: Any | None
    related_users: Any | None
    user: UserCompact | None
    
    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True

# https://osu.ppy.sh/docs/index.html#beatmapset
class Beatmapset(BeatmapsetCompact):
    availability: Availability
    bpm: float
    can_be_hyped: bool
    creator: str
    discussion_locked: bool
    hype: Hype | None
    is_scoreable: bool
    last_updated: datetime
    legacy_thread_url: str | None
    nominations: Nominations | None
    ranked: int
    ranked_date: datetime | None
    source: str
    storyboard: bool
    submitted_date: datetime | None
    tags: str

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        json_encoders = {
            datetime: str,
		}


from pydantic import BaseModel
from .Beatmapset import Beatmapset
from .UserStatistics import UserStatistics

class Spotlight(BaseModel):
    end_date: str
    id: int
    mode_specific: bool
    participant_count: int | None
    name: str
    start_date: str
    type: str

    class Config:
        arbitrary_types_allowed = True

class Spotlights(BaseModel):
    spotlights: list[Spotlight]

    class Config:
        arbitrary_types_allowed = True

class Cursor(BaseModel):
    page: int

class Rankings(BaseModel):
    beatmapsets: list[Beatmapset] | None
    cursor: Cursor | None
    ranking: list[UserStatistics] | None
    spotlight: Spotlight | None
    total: int | None

    class Config:
        arbitrary_types_allowed = True


from datetime import datetime
from pydantic import BaseModel
from .Enums import GameMode, GameModeInt
from .Beatmapset import Beatmapset, BeatmapsetCompact
from .Extras import Failtimes

# https://osu.ppy.sh/docs/index.html#beatmapcompact
class BeatmapCompact(BaseModel):
    beatmapset_id: int
    difficulty_rating: float
    id: int
    mode: GameMode
    status: str
    total_length: int
    user_id: int
    version: str
    
    # optionals
    beatmapset: Beatmapset|BeatmapsetCompact | None
    checksum: str | None
    failtimes: Failtimes | None
    max_combo: int | None

    class Config:
        arbitrary_types_allowed = True
        use_enum_values = True
    
# https://osu.ppy.sh/docs/index.html#beatmap
class Beatmap(BeatmapCompact):
    accuracy: float
    ar: float
    beatmapset_id: int
    bpm: float | None
    convert: bool
    count_circles: int
    count_sliders: int
    count_spinners: int
    cs: float
    deleted_at: datetime | None
    drain: float
    hit_length: int
    is_scoreable: bool
    last_updated: datetime
    mode_int: GameModeInt
    passcount: int
    playcount: int
    ranked: int
    url: str

    class Config:
        arbitrary_types_allowed = True
        use_enum_values = True
        json_encoders = {
            datetime: str,
        }
        
class Beatmaps(BaseModel):
    beatmaps: list[Beatmap]
    
    class Config:
        arbitrary_types_allowed = True
        
class BeatmapPlaycount(BaseModel):
    beatmap_id: int
    beatmap: BeatmapCompact | None
    beatmapset: BeatmapsetCompact | None
    count: int
    
    class Config:
        arbitrary_types_allowed = True


from datetime import datetime
from pydantic import BaseModel, Field
from .User import UserCompact
from .Beatmap import BeatmapCompact
from .Beatmapset import BeatmapsetCompact
from .Extras import Weight, ScoreMatchInfo, Statistics
from .Enums import GameModeInt

# https://osu.ppy.sh/docs/index.html#score
class Score(BaseModel):
    id: int
    best_id: int
    user_id: int
    accuracy: float
    mods: list[str]
    perfect: bool
    statistics: Statistics
    passed: bool
    pp: float
    rank: str
    created_at: datetime
    mode: str
    mode_int: GameModeInt
    replay: bool
    
    # optionals
    beatmap: BeatmapCompact | None
    beatmapset: BeatmapsetCompact | None
    rank_country: int | None
    rank_global: int | None
    weight: Weight | None
    user: UserCompact | None
    score_match: ScoreMatchInfo | None = Field(alias="match")

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        use_enum_values = True
        json_encoders = {
            datetime: str,
		}
        
# https://osu.ppy.sh/docs/index.html#beatmapuserscore
class BeatmapUserScore(BaseModel):
    position: int
    score: Score
    
# https://osu.ppy.sh/docs/index.html#beatmapscores
class BeatmapScores(BaseModel):
    scores: list[Score]
    userScore: BeatmapUserScore | None

    class Config:
        arbitrary_types_allowed = True
        
class Scores(BaseModel):
    scores: list[Score]
    
    class Config:
        arbitrary_types_allowed = True


from datetime import datetime
from typing import Any
from pydantic import BaseModel, Field
from .UserExtras import UserAccountHistory, UserBadge, UserGroup, UserMonthlyPlaycount, UserAchievement, UserProfileCustomization
from .UserStatistics import UserStatistics, UserStatisticsRulesets
from .Extras import ProfileBanner, Country, Cover, Page, RankHighest, RankHistory, ReplaysWatchedCount, Kudosu
from .Enums import GameMode

# https://osu.ppy.sh/docs/index.html#usercompact
class UserCompact(BaseModel):
    avatar_url: str
    country_code: str
    default_group: str
    id: int
    is_active: bool
    is_bot: bool
    is_deleted: bool
    is_online: bool
    is_supporter: bool
    last_vist: datetime | None
    pm_friends_only: bool
    profile_color: str | None = Field(alias="profile_colour")
    username: str

    # https://osu.ppy.sh/docs/index.html#usercompact-optionalattributes
    account_history: list[UserAccountHistory] | None
    active_tournament_banner: ProfileBanner | None
    badges: list[UserBadge] | None
    beatmap_playcounts_count: int | None
    blocks: Any | None
    country: Country | None
    cover: Cover | None
    favorite_beatmapset_count: int | None = Field(alias="favourite_beatmapset_count")
    follower_count: int | None
    graveyard_beatmapset_count: int | None
    groups: list[UserGroup] | None
    is_restricted: bool | None
    loved_beatmapset_count: int | None
    monthly_playcounts: list[UserMonthlyPlaycount] | None
    page: Page | None
    pending_beatmapset_count: int | None
    previous_usernames: list[str] | None
    rank_highest: RankHighest | None
    rank_history: RankHistory | None
    ranked_beatmapset_count: int | None
    replays_watched_counts: list[ReplaysWatchedCount] | None
    scores_best_count: int | None
    scores_first_count: int | None
    scores_recent_count: int | None
    statistics: UserStatistics | None
    statistics_rulesets: UserStatisticsRulesets | None
    support_level: int | None
    unread_pm_count: int | None
    user_achievements: UserAchievement | None
    user_preferences: UserProfileCustomization | None

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        use_enum_values = True
        json_encoders = {
            datetime: str,
		}

# https://osu.ppy.sh/docs/index.html#user
class User(UserCompact):
    cover_url: str
    discord: str | None
    has_supported: bool
    interests: str | None
    join_date: datetime
    kudosu: Kudosu
    location: str | None
    max_blocks: int
    max_friends: int
    occupation: str | None
    playmode: GameMode
    playstyle: list[str]
    post_count: int
    profile_order: list[str]
    title: str | None
    title_url: str | None
    twitter: str | None
    website: str | None

    class Config:
        arbitrary_types_allowed = True
        use_enum_values = True
        json_encoders = {
            datetime: str,
		}
        
class Users(BaseModel):
    users: list[UserCompact]
    
    class Config:
        arbitrary_types_allowed = True


from datetime import datetime
from pydantic import BaseModel
from .Enums import UserAcountHistoryTypes, GameMode, BeatmapsetDownload, UserListFilters, UserListSorts, UserListViews

# https://osu.ppy.sh/docs/index.html#usergroup
class UserGroup(BaseModel):
    playmodes: list[GameMode] | None
    
# https://osu.ppy.sh/docs/index.html#usercompact-useraccounthistory
class UserAccountHistory(BaseModel):
    description: str | None
    id: int
    length: int
    permanent: bool
    timestamp: datetime
    type: UserAcountHistoryTypes
    
    class Config:
        arbitrary_types_allowed = True
        json_encoders = {
            datetime: str,
        }

# https://osu.ppy.sh/docs/index.html#usercompact-userbadge
class UserBadge(BaseModel):
    awarded_at: datetime
    description: str
    image_url: str
    url: str
    
    class Config:
        arbitrary_types_allowed = True
        json_encoders = {
            datetime: str, 
        }
    
class UserMonthlyPlaycount(BaseModel):
    start_date: str
    count: int

class UserAchievement(BaseModel):
    achieved_at: str
    achievement_id: int | None

class UserProfileCustomization(BaseModel):
    audio_autoplay: bool | None
    audio_muted: bool | None
    audio_volume: int | None
    beatmapset_download: BeatmapsetDownload | None
    beatmapset_show_nsfw: bool | None
    beatmapset_title_show_original: bool | None
    comments_show_deleted: bool | None
    forum_posts_show_deleted: bool
    ranking_expanded: bool
    user_list_filter: UserListFilters | None
    user_list_sort: UserListSorts | None
    user_list_view: UserListViews | None

    class Config:
        arbitrary_types_allowed = True


from pydantic import BaseModel, Field

# https://osu.ppy.sh/docs/index.html#beatmapsetcompact-covers
class Covers(BaseModel):
    cover: str
    cover2x: str = Field(alias="cover@2x")
    card: str
    card2x: str = Field(alias="card@2x")
    cover_list: str = Field(alias="list")
    cover_list2x: str = Field(alias="list@2x")
    slimcover: str
    slimcover: str = Field(alias="slimcover@2x")
    
    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True


from functools import wraps
import httpx
from pydantic import parse_obj_as
from .types import Beatmap, Beatmaps, Rankings, User, Scores, Score, GameMode, GameModeInt, RankingType, ScoreTypes, BeatmapUserScore, BeatmapScores, Attributes, KudosuHistory, BeatmapType, BeatmapPlaycount, Beatmapset, UserCompact, Users, Event, Spotlights
from .utility import c_TypeError

class OsuApi:
    base_headers = {
            "Accept" : "application/json",
            "Content-Type" : "application/json",
        }
    
    BASE_URL = "https://osu.ppy.sh/api/v2"
    TOKEN_URL = "https://osu.ppy.sh/oauth/token"

    def __init__ (self, client_id:int, client_secret:str)->None:
        self.Client = httpx.Client()
        self.client_id = client_id
        self.client_secret = client_secret
        self.authorization = self.__new_auth()

    #? https://osu.ppy.sh/docs/index.html#client-credentials-grant
    def __new_auth(self):
        body_params = {
            "client_id" : self.client_id,
            "client_secret" : self.client_secret,
            "grant_type" : "client_credentials",
            "scope" : "public",
        }

        response = self.Client.post(url=self.TOKEN_URL, json=body_params, headers=self.base_headers).json()
        if "error" in response:
            raise ConnectionError(f'error: {response["error"]}')
        return response["token_type"] +" "+ response["access_token"]

    def __del_auth(self):
        if self.authorization == None:
            raise ValueError("self.authorization is type<None> or is not set.")

        headers = self.base_headers
        headers["Authorization"] = self.authorization

        self.Client.delete(url=self.BASE_URL+"/oauth/tokens/current", headers=headers)
        self.authorization = None

    def __del__(self):
        self.__del_auth()
        self.Client.close()
    
    def __exit__(self):
        self.__del_auth()
        self.Client.close()

    def verify_auth(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.authorization:
                self.authorization = self.__new_auth()
            headers = self.base_headers
            headers["Authorization"] = self.authorization
            return func(self, headers=headers, *args, **kwargs)
        return wrapper

    #? https://osu.ppy.sh/docs/index.html#lookup-beatmap
    @verify_auth
    def lookup_beatmap(self,
                       beatmap_id:int, 
                       checksum:str=None, 
                       filename:str=None,
                       **kwargs) -> Beatmap:
        query_params = {}

        if not isinstance(beatmap_id, int):
            raise c_TypeError(param_name="beatmap_id",correct="int",wrong=type(beatmap_id).__name__)
        query_params["id"] = beatmap_id

        if checksum:
            if not isinstance(checksum, str):
                raise c_TypeError(param_name="checksum", correct="str", wrong=type(checksum).__name__)
            query_params["checksum"] = checksum

        if filename:
            if not isinstance(filename, str):
                raise c_TypeError(param_name="filename", correct="str", wrong=type(filename).__name__)
            query_params["filename"] = filename


        response = self.Client.get(url=self.BASE_URL+"/beatmaps/lookup", params=query_params, headers=kwargs["headers"])
        return parse_obj_as(type_=Beatmap, obj=response.json())
    
    #? https://osu.ppy.sh/docs/index.html#get-a-user-beatmap-score
    @verify_auth
    def user_beatmap_score(self,
                           beatmap_id:int, 
                           user_id:int, 
                           mode:GameMode|str=None, 
                           mods:str=None,
                           **kwargs)->BeatmapUserScore:
        query_params = {}
        
        if not isinstance(beatmap_id, int):
            raise c_TypeError(param_name="beatmap_id",correct="int",wrong=type(beatmap_id).__name__)
        if not isinstance(user_id, int):
            raise c_TypeError(param_name="user_id",correct="int",wrong=type(user_id).__name__)
        
        if mode:
            if not isinstance(mode, (GameMode, str)):
                raise c_TypeError(param_name="mode",correct="GameMode|str",wrong=type(mode).__name__)
            if isinstance(mode, GameMode):
                mode = mode.value
            query_params["mode"] = mode
        
        if mods:
            if not isinstance(mods, str):
                raise c_TypeError(param_name="mods",correct="str",wrong=type(mods).__name__)
            query_params["mods"] = mods

        response = self.Client.get(url=self.BASE_URL+f"/beatmaps/{beatmap_id}/scores/users/{user_id}", headers=kwargs["headers"], params=query_params)
        return parse_obj_as(type_=BeatmapUserScore, obj=response.json())
    
    #? https://osu.ppy.sh/docs/index.html#get-a-user-beatmap-scores
    @verify_auth
    def user_beatmap_scores(self,
                            beatmap_id:int, 
                            user_id:int, 
                            mode:GameMode|str=None,
                            **kwargs)->Scores:
        query_params = {}

        if not isinstance(beatmap_id, int):
            raise c_TypeError(param_name="beatmap_id",correct="int",wrong=type(beatmap_id).__name__)
        if not isinstance(user_id, int):
            raise c_TypeError(param_name="user_id",correct="int",wrong=type(user_id).__name__)

        if mode:
            if not isinstance(mode, (GameMode, str)):
                raise c_TypeError(param_name="mode",correct="GameMode|str",wrong=type(mode).__name__)
            if isinstance(mode, GameMode):
                mode = mode.value
            query_params["mode"] = mode
        
        response = self.Client.get(url=self.BASE_URL+f"/beatmaps/{beatmap_id}/scores/users/{user_id}/all", headers=kwargs["headers"], params=query_params)
        return parse_obj_as(type_=Scores, obj=response.json())
    
    #? https://osu.ppy.sh/docs/index.html#get-beatmap-scores
    @verify_auth
    def beatmap_scores(self,
                       beatmap_id:int, 
                       mode:GameMode|str=None, 
                       mods:str=None, 
                       Type:str=None,
                       **kwargs)->BeatmapScores:
        query_params = {}

        if not isinstance(beatmap_id, int):
            raise TypeError("param:beatmap_id must be type<int>")
    
        if mode:
            if not isinstance(mode, (GameMode, str)):
                raise c_TypeError(param_name="mode",correct="GameMode|str",wrong=type(mode).__name__)
            if isinstance(mode, GameMode):
                mode = mode.value
            query_params["mode"] = mode
        
        if mods:
            if not isinstance(mods, str):
                raise c_TypeError(param_name="mods",correct="str",wrong=type(mods).__name__)
            query_params["mods"] = mods

        if type:
            if not isinstance(Type, str):
                raise c_TypeError(param_name="Type",correct="str",wrong=type(Type).__name__)
            
        response = self.Client.get(url=self.BASE_URL+f"/beatmaps/{beatmap_id}/scores", headers=kwargs["headers"], params=query_params)
        return parse_obj_as(type_=BeatmapScores, obj=response.json())
    
    #? https://osu.ppy.sh/docs/index.html#get-beatmaps
    @verify_auth
    def beatmaps(self,
                 beatmap_ids:list[int],
                 **kwargs)->Beatmaps:
        query_params = {}

        if not isinstance(beatmap_ids, list):
            raise c_TypeError(param_name="beatmap_ids", correct="list", wrong=type(beatmap_ids).__name__)
        if not isinstance(beatmap_ids[0],int):
            raise c_TypeError(param_name="beatmap_ids[elements]", correct="int", wrong=type(beatmap_ids[0]).__name__)
        query_params["ids[]"] = beatmap_ids

        response = self.Client.get(url=self.BASE_URL+"/beatmaps", headers=kwargs["headers"], params=query_params)
        return parse_obj_as(type_=Beatmaps, obj=response.json())
    
    #? https://osu.ppy.sh/docs/index.html#get-beatmap
    @verify_auth
    def beatmap(self,
                beatmap_id:int,
                **kwargs)->Beatmap:

        if not isinstance(beatmap_id, int):
            raise c_TypeError(param_name="beatmap_id", correct="int", wrong=type(beatmap_id).__name__)
        
        response = self.Client.get(url=self.BASE_URL+f"/beatmaps/{beatmap_id}", headers=kwargs["headers"])
        return parse_obj_as(type_=Beatmap, obj=response.json())
    
    #? https://osu.ppy.sh/docs/index.html#get-beatmap-attributes
    @verify_auth
    def beatmap_attributes(self,
                           beatmap_id:int,
                           mods:list[str]=None,
                           ruleset: GameMode|str=None,
                           ruleset_id: GameModeInt|int=None,
                           **kwargs)->Attributes:
        query_params = {}

        if not isinstance(beatmap_id, int):
            raise c_TypeError(param_name="beatmap_id", correct="int", wrong=type(beatmap_id).__name__)
        
        if mods:
            if not isinstance(mods, list):
                raise c_TypeError(param_name="mods", correct="list", wrong=type(mods).__name__)
            if not isinstance(mods[0],int):
                raise c_TypeError(param_name="mods[elements]", correct="str", wrong=type(mods[0]).__name__)
            query_params["mods"] = mods

        if ruleset:
            if not isinstance(ruleset, (GameMode, str)):
                raise c_TypeError(param_name="ruleset", correct="GameMode|str",wrong=type(ruleset).__name__)
            if isinstance(ruleset,GameMode):
                ruleset = ruleset.value
            query_params["ruleset"] = ruleset
        
        if ruleset_id:
            if not isinstance(ruleset_id, (GameModeInt, int)):
                raise c_TypeError(param_name="ruleset_id", correct="GameMode|int",wrong=type(ruleset_id).__name__)
            query_params["ruleset_id"] = ruleset_id

        response = self.Client.post(url=self.BASE_URL+f"/beatmaps/{beatmap_id}/attributes", headers=kwargs["headers"], params=query_params)
        return parse_obj_as(type_=Attributes, obj=response.json())
    
    #? https://osu.ppy.sh/docs/index.html#get-user-kudosu
    @verify_auth
    def user_kudosu(self,
                    user_id:int, 
                    limit:int=None, 
                    offset:str=None,
                    **kwargs)->list[KudosuHistory]:
        query_params = {}

        if not isinstance(user_id, int):
            raise c_TypeError(param_name="user_id", correct="int", wrong=type(user_id).__name__)
        
        if limit:
            if not isinstance(limit, int):
                raise c_TypeError(param_name="limit", correct="int", wrong=type(limit).__name__)
            query_params["limit"] = limit

        if offset:
            if not isinstance(offset, str):
                raise c_TypeError(param_name="offset", correct="str", wrong=type(offset).__name__)
            query_params["offset"] = offset
            
        response = self.Client.get(url=self.BASE_URL+f"/users/{user_id}/kudosu", headers=kwargs["headers"], params=query_params)
        return parse_obj_as(type_=list[KudosuHistory], obj=response.json())
    
    #? https://osu.ppy.sh/docs/index.html#get-user-scores
    @verify_auth
    def user_scores(self,
                    user_id:int, 
                    Type:str, 
                    include_fails:int=0,
                    mode:GameMode|str=None,
                    limit:int=None,
                    offset:int=None,
                    **kwargs)->list[Score]:
        query_params = {}
        
        if include_fails != 0:
            if Type != ScoreTypes.RECENT.value:
                raise ValueError("type must be 'recent' in order to set include_fails")
            elif include_fails not in (0,1):
                raise ValueError("include_fails can only be either 1 or 0")
            query_params["include_fails"] = include_fails

        if mode:
            if not isinstance(mode, (GameMode, str)):
                raise c_TypeError(param_name="mode",correct="GameMode|str",wrong=type(mode).__name__)
            if isinstance(mode, GameMode):
                mode = mode.value
            query_params["mode"] = mode

        if limit:
            if not isinstance(limit, int):
                raise c_TypeError(param_name="limit",correct="int",wrong=type(limit).__name__)  
            query_params["limit"] = limit
            
        if offset:
            if not isinstance(offset, int):
                raise c_TypeError(param_name="offset",correct="int",wrong=type(offset).__name__) 
            query_params["offset"] = offset

        response = self.Client.get(url=self.BASE_URL+f"/users/{user_id}/scores/{type}", headers=self.base_headers, params=query_params)
        return parse_obj_as(type_=list[Score], obj=response.json())
                
    #? https://osu.ppy.sh/docs/index.html#get-user-beatmaps
    @verify_auth
    def user_beatmaps(self,
                      user_id:int,
                      Type:BeatmapType|str,
                      limit:int=None,
                      offset:str=None,
                      **kwargs)->list[BeatmapPlaycount] | list[Beatmapset]:
        query_params = {}
        
        if not isinstance(user_id, int):
            raise c_TypeError(param_name="user_id", correct="int", wrong=type(user_id).__name__)
        if not isinstance(Type, (BeatmapType,str)):
            raise c_TypeError(param_name="Type", correct="BeatmapType|str", wrong=type(Type).__name__)  
        if isinstance(Type, BeatmapType):
            Type = Type.value
         
        if limit:
            if not isinstance(limit, int):
                raise c_TypeError(param_name="limit", correct="int", wrong=type(limit).__name__)
            query_params["limit"] = limit
        
        if offset:
            if not isinstance(offset, str):
                raise c_TypeError(param_name="offset", correct="str", wrong=type(offset).__name__)
            query_params["offset"] = offset
            
        response = self.Client.get(url=self.BASE_URL+f"/users/{user_id}/beatmapsets/{Type}", headers=kwargs["headers"], params=query_params)
        if Type == BeatmapType.MOST_PLAYED.value:
            return parse_obj_as(type_=list[BeatmapPlaycount], obj=response.json())
        return parse_obj_as(type_=list[Beatmapset], obj=response.json())
    
    #? https://osu.ppy.sh/docs/index.html#get-user-recent-activity
    @verify_auth
    def user_recent_activity(self,
                             user_id:int,
                             limit:int=None,
                             offset:str=None,
                             **kwargs)->list[Event]:
        query_params = {}
        
        if not isinstance(user_id, int):
            raise c_TypeError(param_name="user_id", correct="int", wrong=type(user_id).__name__)
         
        if limit:
            if not isinstance(limit, int):
                raise c_TypeError(param_name="limit", correct="int", wrong=type(limit).__name__)
            query_params["limit"] = limit
        
        if offset:
            if not isinstance(offset, str):
                raise c_TypeError(param_name="offset", correct="str", wrong=type(offset).__name__)
            query_params["offset"] = offset

        response = self.Client.get(url=self.BASE_URL+f"/users/{user_id}/recent_activity", headers=kwargs["headers"], params=query_params)
        return parse_obj_as(type_=list[Event], obj=response.json())
    
    #? https://osu.ppy.sh/docs/index.html#get-user
    @verify_auth
    def user(self,
             username:int|str,
             mode:GameMode|str="",
             key:str="",
             **kwargs)-> User:
        query_params = {}

        if not isinstance(username, (int,str)):
            raise c_TypeError(param_name="username",correct="int|str",wrong=type(username).__name__)

        if mode != "":
            if not isinstance(mode, (GameMode, str)):
                raise c_TypeError(param_name="mode",correct="GameMode|str",wrong=type(mode).__name__)
            if isinstance(mode, GameMode):
                mode = mode.value

        if key != "":
            if key not in ["id", "username"]:
                raise ValueError("key can only be 'id' or 'username'")
            query_params["key"] = key

        response = self.Client.get(url=self.BASE_URL+f"/users/{username}/{mode}", headers=kwargs["headers"],params=query_params)
        return parse_obj_as(type_=User, obj=response.json())

    #? https://osu.ppy.sh/docs/index.html#get-users
    @verify_auth
    def users(self,
              user_ids:list[int],
              **kwargs)->Users:
        query_params = {}

        if not isinstance(user_ids, list):
            raise c_TypeError(param_name="user_ids", correct="list", wrong=type(user_ids).__name__)
        if not isinstance(user_ids[0], int):
            raise c_TypeError(param_name="user_ids[elements]", correct="int", wrong=type(user_ids[0]).__name__)
        query_params["ids[]"] = user_ids

        response = self.Client.get(url=self.BASE_URL+"/users", headers=kwargs["headers"], params=query_params)
        return parse_obj_as(type_=Users, obj=response.json())

    #? https://osu.ppy.sh/docs/index.html#get-ranking
    @verify_auth
    def ranking(self,
                mode:GameMode|str,
                Type:str,
                filter:str="all",
                country:int=None,
                cursor:int=None,
                spotlight_id:int=None,
                variant:str=None,
                **kwargs)->Rankings:
        query_params = {}

        if not isinstance(mode, (GameMode, str)):
            raise c_TypeError(param_name="mode",correct="GameMode|str",wrong=type(mode).__name__)
        if isinstance(mode, GameMode):
            mode = mode.value

        if not isinstance(Type, (RankingType, str)):
            raise c_TypeError(param_name="Type", correct="RankingType|str", wrong=type(Type).__name__)
        if isinstance(Type, RankingType):
            Type = Type.value

        if filter not in ["all","friends"]:
            raise ValueError("filter must be 'all' or 'friends'")
        query_params["filter"] = filter

        if cursor:
            if not isinstance(cursor, int):
                raise c_TypeError(param_name="cursor",correct="int",wrong=type(cursor).__name__)
            if cursor < 0:
                raise ValueError("cursor must be greater that -1")    
            query_params["cursor"] = cursor
        
        if country:
            if not isinstance(country, int):
                raise c_TypeError(param_name="country",correct="int",wrong=type(country).__name__)
            if type != RankingType.PERFORMANCE.value:
                raise ValueError("type must be 'performance' in order to set a country code.")
            query_params["country"] = country

        if spotlight_id:
            if type != RankingType.CHARTS.value:
                raise ValueError("type must be 'charts' in order to set a spotlight_id.")
            query_params["spotlight"] = spotlight_id

        if variant:
            if variant not in ["4k","7k"]:
                raise ValueError("variant can only be '4k' or '7k'")
            if type != RankingType.PERFORMANCE.value:
                raise ValueError("type must be 'performance' to use variant")
            if mode != GameMode.MANIA.value:
                raise ValueError("mode must be 'mania' to use variant")
            query_params["variant"] = variant

        response = self.Client.get(url=self.BASE_URL+f"/rankings/{mode}/{Type}", headers=kwargs["headers"], params=query_params)
        return parse_obj_as(type_=Rankings, obj=response.json())
    
    #? https://osu.ppy.sh/docs/index.html#get-spotlights
    @verify_auth
    def spotlights(self, **kwargs)->Spotlights:

        response = self.Client.get(url=self.BASE_URL+"/spotlights", headers=kwargs["headers"])
        return parse_obj_as(type_=Spotlights, obj=response.json())


from enum import IntEnum, Enum

class ExtendedEnum(Enum):

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))

class GameMode(ExtendedEnum):
    OSU = "osu"
    TAIKO = "taiko"
    FRUITS = "fruits"
    MANIA = "mania"

class GameModeInt(IntEnum):
    OSU = 0
    TAIKO = 1
    FRUITS = 2
    MANIA = 3

class RankingType(ExtendedEnum):
    CHARTS = "charts"
    COUNTRY = "country"
    PERFORMANCE = "performance"
    SCORE = "score"

class ScoreTypes(ExtendedEnum):
    BEST = "best"
    FIRSTS = "firsts"
    RECENT = "recent"
    
class UserAcountHistoryTypes(ExtendedEnum):
    NOTE = "note"
    RESTRICTION = "restriction"
    SILENCE = "silence"
    
class BeatmapType(ExtendedEnum):
    FAVORITE = "favourite"
    GRAVEYARD = "graveyard"
    LOVED = "loved"
    MOST_PLAYED = "most_played"
    PENDING = "pending"
    RANKED = "ranked"
    
class BeatmapsetDownload(ExtendedEnum):
    ALL = "all"
    NO_VIDEO = "no_video"
    DIRECT = "direct"

class UserListFilters(ExtendedEnum):
    ALL = "all"
    ONLINE = "online"
    OFFLINE = "offline"
    
class UserListSorts(ExtendedEnum):
    LAST_VISIT = "last_visit"
    RANK = "rank"
    USERNAME = "username"
    
class UserListViews(ExtendedEnum):
    CARD = "card"
    LIST = "list"
    BRICK = "brick"


from datetime import datetime	
from pydantic import BaseModel
from .Extras import Achievement
from .Enums import GameMode

class EventBeatmap(BaseModel):
	title: str
	url: str

class EventBeatmapset(BaseModel):
	title: str
	url: str

class EventUser(BaseModel):
	username: str
	url: str
	previousUsername: str | None

class Event(BaseModel):
	created_at: datetime
	id: int
	type: str
	achievement: Achievement | None
	user: EventUser | None
	beatmap: EventBeatmap | None
	count: int | None
	approval: str | None
	beatmapset: EventBeatmapset | None
	scoreRank: str | None
	rank: int | None
	mode: GameMode | None

	class Config:
		arbitrary_types_allowed = True
		use_enum_values = True
		json_encoders = {
			datetime: str
		}


from pydantic import BaseModel, Field
from .DifficultyAttributes import DifficultyAttributes

class GradeCounts(BaseModel):
    a:int
    s:int
    sh:int
    ss:int
    ssh:int

class Level(BaseModel):
    current: int
    progress: int
    
class Cover(BaseModel):
    custom_url: str | None
    url: str
    id: int | None

class Country(BaseModel):
    code: str
    name: str

class RankHistory(BaseModel):
    mode: str
    data: list[int]

# https://osu.ppy.sh/docs/index.html#usercompact-profilebanner
class ProfileBanner(BaseModel):
    id: int
    tournament_id: int
    image: str
    
# https://osu.ppy.sh/docs/index.html#usercompact-rankhighest
class RankHighest(BaseModel):
    rank: int
    updated_at: str
    
class ReplaysWatchedCount(BaseModel):
    start_date: str
    count: int
    
    
class Availability(BaseModel):
    download_disabled: bool
    more_information: str | None
    
class Hype(BaseModel):
    current: int
    required: int

class Nominations(BaseModel):
    current: int
    required: int
    
class Weight(BaseModel):
    percentage: float
    pp: int

class Page(BaseModel):
    html: str
    raw: str

class Kudosu(BaseModel):
    total: int
    available: int

class Statistics(BaseModel):
    count_50: int
    count_100: int
    count_300: int
    count_geki: int
    count_katu: int
    count_miss: int

class ScoreMatchInfo(BaseModel):
    slot: int
    team: str
    score_match_pass: bool = Field(alias="pass")
    
class Attributes(BaseModel):
    attributes: DifficultyAttributes

    class Config:
        arbitrary_types_allowed = True

# https://osu.ppy.sh/docs/index.html#kudosuhistory
class Giver(BaseModel):
    url: str
    username: str

# https://osu.ppy.sh/docs/index.html#kudosuhistory
class Post(BaseModel):
    url: str | None
    title: str
    
class Failtimes(BaseModel):
    exit: list[int] | None
    fail: list[int] | None
    
class Achievement(BaseModel):
	achieved_at: str
	achievement_id: int


from pydantic import BaseModel
from .Extras import Level, GradeCounts

# https://osu.ppy.sh/docs/index.html#userstatistics
class UserStatistics(BaseModel):
    grade_counts: GradeCounts
    hit_accuracy: int
    is_ranked: bool
    level: Level
    maximum_combo: int
    play_count: int
    play_time: int
    pp: int
    global_rank: int | None
    ranked_score: int
    replays_watched_by_others: int
    total_hits: int
    total_score: int

    class Config:
        arbitrary_types_allowed = True
        use_enum_values = True

class UserStatisticsRulesets(BaseModel):
    osu: UserStatistics | None
    mania: UserStatistics | None
    taiko: UserStatistics | None
    fruits: UserStatistics | None

    class Config:
        arbitrary_types_allowed = True


from setuptools import find_packages, setup

with open("README.md", "r") as readme:
    long_description = readme.read()

VERSION = "0.0.7"

setup(
    name="losuapi",
    version=VERSION,
    description="Python wrapper for the Osu apiv2.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["python","osu","api","wrapper"],
    author="Darien Moore",
    author_email="LiskIsBest@gmail.com",
    url="https://github.com/LiskIsBest/Losuapi",
    download_url="https://github.com/LiskIsBest/Losuapi/tarball/v"+VERSION,
    license="GPLv3",
    packages=find_packages(),
    install_requires=[
		"httpx",
        "pydantic"
	]
)


from .Beatmap import Beatmap, BeatmapCompact, Beatmaps, BeatmapPlaycount
from .Beatmapset import Beatmapset, BeatmapsetCompact
from .Covers import Covers
from .DifficultyAttributes import DifficultyAttributes
from .Enums import UserAcountHistoryTypes, UserListFilters, UserListSorts, UserListViews, GameMode, GameModeInt, ScoreTypes, RankingType, BeatmapsetDownload, BeatmapType
from .Extras import GradeCounts, Level, Cover, Country, RankHighest, RankHistory, ReplaysWatchedCount, ProfileBanner, Availability, Hype, Nominations, Weight, Page, Kudosu, Statistics, ScoreMatchInfo, Attributes, Giver, Post, Failtimes
from .KudosuHistory import KudosuHistory
from .Rankings import Spotlight, Spotlights, Cursor, Rankings
from .Score import Score, BeatmapScores, BeatmapUserScore, Scores
from .User import User, UserCompact, Users
from .UserExtras import UserGroup, UserAccountHistory, UserBadge, UserMonthlyPlaycount, UserAchievement, UserProfileCustomization
from .UserStatistics import UserStatistics, UserStatisticsRulesets
from .Event import Event


from pydantic import BaseModel, Field

# https://osu.ppy.sh/docs/index.html#beatmapdifficultyattributes
class DifficultyAttributes(BaseModel):
    max_combo: int
    star_rating: float
    aim_difficulty: float | None
    approach_rate: float | None
    flastlight_difficulty: float | None
    overall_difficulty: float | None
    slider_factor: float | None
    speed_difficulty: float | None
    stamina_difficulty: float | None
    rhythm_difficulty: float | None
    color_difficulty: float | None = Field(alias="colour_difficulty")
    great_hit_window: float | None
    score_multiplier: float | None

    class Config:
        allow_population_by_field_name = True
        
class Attributes(BaseModel):
    attributes: DifficultyAttributes
    
    class Config:
        arbitrary_types_allowed = True